#!/usr/bin/perl
use warnings;
use strict;

use IO::File;

use constant DEBUG => 0;

{
	package PG::TocEntry;

	use Carp qw(croak);

	use constant DEBUG => 0;

	sub new {
		my $class    = shift;
		my $self     = {};
		$self->{label}     = '';
		$self->{href}      = '';
		$self->{as_text}   = '';
		$self->{child_toc} = undef;

		bless ($self,$class);
		return $self;
	}

	sub new_from_str {
		my ($class, $str) = @_;
		my $self = new($class);
		$self->parse($str);

		return $self;
	}

	# Returns true if link is to a whole page rather than an anchor within a page
	#
	sub is_full_page {
		my $self = shift;
		return !($self->{href} =~ /html#/);
	}

	sub matches_sub_toc {
		my ($self, $toc_name) = @_;

		if ($self->{href} eq $toc_name) { return $self; }
		if ($self->{child_toc}) { return $self->{child_toc}->matches_sub_toc($toc_name); }
		return 0;
	}

	sub insert_sub_toc {
		my ($self, $child_toc) = @_;
		###croak "Already have child-toc" if ($self->{child_toc});
		$self->{child_toc} = $child_toc;
	}

	sub parse {
		my $self = shift;
		my $str  = shift;

		if (DEBUG) { print STDERR "TocEntry: ".substr($str,0,40)."... "; }

		# We expect something like <DT> <A HREF="...">...</A></DT>
		# If following that is <DD> look for a sub-tree
		unless ($str =~ m{\A \s* <DT> [\w\.\s]* <A \s+ HREF="([^"]+)" > (.+?) </A> .*? </DT>}xms) {
			print STDERR "\n";
			if (!$2) { print STDERR "match 2 undefined\n"; }
			if (!$1) { print STDERR "match 1 undefined\n"; }
			die "Unable to parse: ".substr($str, 0, 200)."...";
		}
		$self->{href}    = $1;
		$self->{label}   = $2;
		$self->{as_text} = $&;

		$str = substr($str, length($self->{as_text}));
		if ($str =~ m{\A <DD> <DL>}xms) {
			if (DEBUG) { print STDERR "Child!\n"; }
			my $child_toc = PG::Toc->new_from_str($str);
			if ($child_toc->not_empty) {
				$self->{child_toc} = $child_toc;
			}
			my $len = length('<DD>') + $child_toc->{len} + length('</DD>');
			if (DEBUG) { print STDERR "subtoc len = $len\n"; }
			$self->{as_text} .= substr($str, 0, $len);
		}

		if (DEBUG) { print STDERR "len=".$self->raw_text_len()."\n"; }
	}

	sub raw_text_len {
		my $self = shift;
		return length($self->{as_text});
	}

	sub print_to {
		my $self = shift;
		my $fh   = shift;
		my $lvl  = shift || 0;

		my $css_class = ($self->{child_toc}) ? ' CLASS="toggle"' : '';
		if ($self->{child_toc}) {
			my $arrow = '<a class="toggle" href="#" onclick="return toggleTree(this)"><img src="rt.png"></a>';
			my $link = $self->_mk_link($self->{label}, $self->{href}, $self->{label});
			print $fh '<DT'.$css_class.'>'.$arrow.$link.'</DT>';
			print $fh '<DD style="display:none">';
			$self->{child_toc}->print_to($fh, $lvl + 1);
			print $fh '</DD>';
		}
		else {
			my $link = $self->_mk_link($self->{label}, $self->{href}, $self->{label});
			print $fh '<DT'.$css_class.'>'.$link.'</DT>';
		}
	}

	sub _mk_link {
		my ($self, $title, $href, $text, $css_class) = @_;

		my $css_attr = defined($css_class) ? qq[class="$css_class"] : '';
		# Strip any HTML tags that have crept into our title attribute
		$title =~ s/<[^>]*?>//g;
		return qq[<a href="$href" title="$title" $css_attr>$text</a>];
	}

	1;
}

{
	package PG::Toc;

	use Carp qw(croak);

	use constant DEBUG => 0;

	sub new {
		my $class    = shift;
		my $self     = {};
		$self->{filename}      = '';
		$self->{nodes}         = [];
		$self->{len}           = 0;

		bless ($self, $class);
		return $self;
	}

	sub not_empty {
		my $self = shift;
		return scalar @{$self->{nodes}};
	}

	sub new_from_str {
		my ($class, $str) = @_;
		my $self = new( $class );
		$self->parse($str);

		return $self;
	}

	sub read {
		my ($self, $filename) = @_;
		if ($filename) { $self->{filename} = $filename;	}

		IO::File->input_record_separator(undef);
		my $fh = IO::File->new($self->{filename}, 'r')
			or croak "Unable to close file: ".$self->{filename}.": $!";
		my $contents  = <$fh>;
		$fh->close or croak "Unable to close file: ".$self->{filename}.": $!";

		$self->parse_full_file($contents);
	}

	sub parse_full_file {
		my $self = shift;
		my $contents = shift;

		# Strip header and footer
		$contents =~ s{\A .+ <DIV \s+ CLASS="TOC" \s* >}{}xms;
		$contents =~ s{</DL \s* > </DIV \s* > </DIV \s* > <DIV \s+ CLASS="NAVFOOTER" .+ \z}{</DL>}xms;

		# Strip intro
		$contents =~ s{<DT\s*><B\s*>Table\sof\sContents</B \s*></DT \s* >}{}xms;
		# Strip chapter numbers
		$contents =~ s{<DT\s*> \w+ \. \s<A}{<DT><A}xmsg;

		# Strip any remaining irritating newlines
		$contents =~ s{\n>}{>}xmsg;
		$contents =~ s{<A\n}{<A }xmsg;

		$self->parse($contents);
	}

	sub parse {
		my $self = shift;
		my $contents = shift;

		if (DEBUG) { print STDERR "Toc:      ".substr($contents,0,40)."...\n"; }

		# Strip anything up to the first <DL>
		$contents =~ s/\A .*? <DL>/<DL>/xms;
		# Strip out the first <DT> if it's a heading.
		$contents =~ s{\A <DT> <B> .+? </B> </DT>}{}xms;

		# Now scan - each <DT> is a node and it knows how to handle sub-trees
		my $len = length($contents);
		my $pos = 4;	# skip <DL>
		while (($len - $pos) > 4) {
			my $node = PG::TocEntry->new_from_str( substr($contents, $pos) );
			$pos += $node->raw_text_len();
			if ($node->is_full_page) {
				push @{$self->{nodes}}, $node;
			}
			if (substr($contents, $pos, 5) eq '</DL>') {
				$pos+=5;
				last;
			}
		}
		if (DEBUG) {
			print STDERR "  end of toc = ",substr($contents, $pos-10, 10),"\n";
			print STDERR "  next bit   = ",substr($contents, $pos, 40),"\n";
		}
		$self->{len} = $pos;
		return $pos;
	}

	sub matches_sub_toc {
		my ($self, $toc_name) = @_;

		for my $n (@{$self->{nodes}}) {
			if (my $tgt = $n->matches_sub_toc($toc_name)) { return $tgt; }
		}
		return 0;
	}

	sub insert_sub_toc {
		my $self     = shift;
		my $subtoc   = shift;
		my $toc_name = shift;

		for my $n (@{$self->{nodes}}) {
			if (my $tgt = $n->matches_sub_toc($toc_name)) {
				if (DEBUG) { print STDERR "sub-toc inserted into ".$tgt->{href}." / ".$tgt->{label}."\n"; }
				$tgt->insert_sub_toc($subtoc);
				last;
			}
		}
	}

	sub print_to_file {
		my ($self, $filename) = @_;
		my $fh = IO::File->new($filename, 'w')
			or die "Unable to open $filename for writing: $!";

		$self->print_to($fh);

		$fh->close
			or die "Unable to close $filename: $!";
	}

	sub print_to {
		my $self = shift;
		my $fh   = shift;
		my $lvl  = shift || 0;

		print $fh '<DL>';
		for my $n (@{$self->{nodes}}) {
			$n->print_to($fh);
		}
		print $fh "</DL>\n\n";
	}

	1;
}

my @toc_files = get_files_with_toc();

my $global_toc = PG::Toc->new;
$global_toc->read('index.html');

for my $tf (@toc_files) {
	next if ($tf eq 'index.html');
	if (DEBUG) { print STDERR "*** file = $tf ***\n"; }
	my $file_toc = PG::Toc->new;
	$file_toc->read($tf);
	$global_toc->insert_sub_toc($file_toc, $tf);
}

$global_toc->print_to_file('contentstree.dat');
exit;


sub get_files_with_toc {
	open(my $fh, q[grep -l 'CLASS="TOC"' *.html |])
		or die "Unable to run grep: $!";
	my @files = map { chomp; $_; } <$fh>;
	close($fh) or die "Unable to close grep: $!";
	return @files;
}
